#include "bkey.h"

/* 
   BotikKey client implementation in plain C
   with command line interface.
   (c)2010 Botik-Technologies Ltd. http://www.botik.ru
   Distributed under the terms of the BSD open-source license.
   See license.txt for details.
*/

char *mode;         /* network access mode */
char *logfile=NULL; /* log file path */
int verbose=0;      /* whether to print send/recv messages */
int quiet=0;        /* whether to do not print anything */
int coloron=0;      /* whether to colour std.output */

/*******************************************/
void print_usage() {
    color_on(Yellow);
    printf("BotikKey command line client v."VER"\n"
            "(c)2010 Botik-Technologies Ltd. http://www.botik.ru\n"
            "Valid command line switches are:\n"
            "  -f <pswd file>: path to file containing BotikKey password in plain text;\n"
            "  -s <server address>: IP-address or hostname of the nearest BotikKey server;\n"
            "  -m <access mode>: network access mode (WORLD,LOCAL or NONE);\n"
            "  -d <pswd digest>: password MD5 digest string in plain text;\n");
    printf( "  -l <log file>: path to logfile (will be created) to append messages to;\n"
            "  -v: turn on verbose output (print send/recv messages);\n"
            "  -c: turn on coloured output;\n"
            "  -q: be quiet, do not print messages to console (excluding password prompt);\n"
#ifdef _MAC
            "  -k: use Keychain to retrieve BotikKey password from;\n"
            "  -K: reset BotikKey password in Keychain;\n"
#endif
            "  -h: show this help.\n");
    color_off();
}
/*******************************************/
int create_socket(int nonblock) {
    int sock, on=1, r, timeo=SOCK_TIMEO;
#ifdef _WIN32
    int tv=timeo;
#else
    struct timeval tv;

    tv.tv_sec=timeo/1000;
    tv.tv_usec=(timeo%1000)*1000;
#endif
    sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (sock < 0) {
        print_error("open socket",get_last_neterr());
        return -1;
    }
    r = setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char*)&on, sizeof(on));
    if (r < 0) {
        print_error("setsockopt (TCP_NODELAY)",get_last_neterr());
        return -1;
    }
    r = ioctlsocket(sock,FIONBIO,(unsigned long*)&nonblock);
    if (r < 0) {
        print_error("ioctlsocket FIONBIO",get_last_neterr());
        return -1;
    }
    r = setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE,(char*)&on,sizeof(on));
    if (r < 0) {
        print_error("setsockopt SO_KEEPALIVE",get_last_neterr());
        return -1;
    }
    r = setsockopt(sock, SOL_SOCKET, SO_SNDTIMEO,(char*)&tv,sizeof(tv));
    if (r < 0) {
        print_error("setsockopt SO_SNDTIMEO",get_last_neterr());
        return -1;
    }
    r = setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO,(char*)&tv,sizeof(tv));
    if (r < 0) {
        print_error("setsockopt SO_RCVTIMEO",get_last_neterr());
        return -1;
    }
    return sock;
}
/*******************************************/
int invalid_option (char opt) {
    return (strchr("cdfhkKlmqsv",opt)==NULL);
}
/*******************************************/
int local_server() {
    struct sockaddr_in addr;
    int s, r;

    s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (s < 0) return 0;
    memset(&addr, 0, sizeof(struct sockaddr_in));
    addr.sin_addr.s_addr = htonl(LHOSTX);
    addr.sin_port = htons(BKPORT);
    addr.sin_family = AF_INET;
    r = bind(s, (struct sockaddr *)&addr, sizeof(struct sockaddr_in));
    return (r < 0);
}
/*******************************************/
int resolve_host(struct sockaddr_in *destaddr, const char *host) {
    struct hostent *hp;
    unsigned int addr=INADDR_NONE;

    destaddr->sin_addr.s_addr = 0;
    destaddr->sin_family = AF_INET;
    hp = gethostbyname(host);
    if (hp == NULL) {
        addr = inet_addr(host);
    }
    if ((hp == NULL) && (addr == INADDR_NONE)) {
        return 1; /* couldn't resolve */
    }
    if (hp != NULL) {
        memcpy((void*)&(destaddr->sin_addr),hp->h_addr,hp->h_length);
        destaddr->sin_family = (unsigned short)hp->h_addrtype;
    } else {
        destaddr->sin_addr.s_addr = addr;
        destaddr->sin_family = AF_INET;
    }
    return 0;
}
/*******************************************/
int connect_to(const char *nodeName, int port) {
    struct sockaddr_in serv_addr;
    int sock, conn, r;

    memset(&serv_addr, 0, sizeof(struct sockaddr_in));
    if (resolve_host(&serv_addr,nodeName)) {
        myprintf(Red,"unable to resolve host %s\n",nodeName);
        return -1;
    }
    sock = create_socket(1);
    serv_addr.sin_port=htons((unsigned short)port);
    if (verbose) myprintf(Yellow,"connecting to %s:%d...\n",nodeName,port);
    conn = connect(sock, (saddr *)&serv_addr, sizeof(struct sockaddr_in));
    if (conn < 0) {
        r = get_last_neterr();
        if (r != WOULDBLOCK && r != INPROGRESS) {
            print_error("connect",r);
            return -1;
        }
    }
    return sock;
}
/*******************************************/
void trim(char *str) {
    char *chr;
    while ((chr = strchr(str,'\r')) != NULL) {
        chr[0] = '\0';
    }
    while ((chr = strchr(str,'\n')) != NULL) {
        chr[0] = '\0';
    }
}
/*******************************************/
void get_response(char *challenge, char *dpassword, char *result) {
    int rnd;
    int hashlen=2*MD5_HASHBYTES+1;
    uchar *string, *digest;

    rnd = myrand();
    string = new_ustring(MSGSIZE);
    digest = new_ustring(hashlen);
    _snprintf((char*)string,MSGSIZE,"%s%d%s%s",challenge,rnd,mode,dpassword);
    make_digest(string,digest);
    _snprintf(result,MSGSIZE,"RESPONSE %s %d %s\r\n",(char*)digest,rnd,mode);
    free(string);
    free(digest);
}
/**************** read the 1st line in file ***************************/
int read_line(char *path, char *pwd, int count) {
    int fd, r;

    fd = _open(path, _O_RDONLY);
    if (fd < 0) {
        myprintf(Red,"Error opening file: %s\n",strerror(errno));
        return 1;
    }
    r = _read(fd,pwd,count);
    if (r < 0) {
        myprintf(Red,"Error reading file: %s\n",strerror(errno));
        return 1;
    }
    trim(pwd);
    r = _close(fd);
    if (r < 0) {
        myprintf(Red,"Error closing file: %s\n",strerror(errno));
    }
    return 0;
}
/************* read the cmdline ******************************/
int parse_args(int argc, char **argv, char *server, int srvlen, ucharp dpassword, int hashlen) {
    int i, kchain=0, Kchain=0, pwlen=64;
    char *pwdfile=NULL;
    ucharp password=NULL;

    for (i = 1; i < argc; i++) {
        if (argv[i][0] != '-' || invalid_option(argv[i][1])) {
            myprintf(Red,"Invalid option: %s\n",argv[i]);
            print_usage();
            return 1;
        }
        if (argv[i][1] == 'f') {
            if (i + 1 < argc) 
                pwdfile = _strdup(argv[++i]);
            else {
                myprintf(Red,"Value must be supplied for option -f.\n");
                return 1;
            }
            continue;
        }
        if (argv[i][1] == 'l') {
            if (i + 1 < argc) 
                logfile = _strdup(argv[++i]);
            else {
                myprintf(Red,"Value must be supplied for option -l.\n");
                return 1;
            }
            continue;
        }
        if (argv[i][1] == 's') {
            if (i + 1 < argc)  {
                strncpy(server,argv[++i],srvlen);
            } else {
                myprintf(Red,"Value must be supplied for option -s.\n");
                return 1;
            }
            continue;
        }
        if (argv[i][1] == 'd') {
            if (i + 1 < argc) 
                strncpy((char*)dpassword,argv[++i],hashlen);
            else {
                myprintf(Red,"Value must be supplied for option -d.\n");
                return 1;
            }
            continue;
        }
        if (argv[i][1] == 'm') {
            if (i + 1 < argc) 
                mode = _strdup(argv[++i]);
            else {
                myprintf(Red,"Value must be supplied for option -m.\n");
                return 1;
            }
            continue;
        }
        if (argv[i][1] == 'v') {
            verbose=1;
            continue;
        }
        if (argv[i][1] == 'c') {
            coloron=1;
            init_colors();
            continue;
        }
        if (argv[i][1] == 'q') {
            quiet=1;
            continue;
        }
        if (argv[i][1] == 'h') {
            print_usage();
            exit_app(EXIT_SUCCESS);
        }
        if (argv[i][1] == 'k') {
            kchain=1;
            continue;
        }
        if (argv[i][1] == 'K') {
            Kchain=1;
            continue;
        }
    }
    if (kchain && Kchain) {
        myprintf(Red,"-k and -K switches should not come together.\n");
        exit_app(EXIT_FAILURE);
    }
    if (init_library()) {
        print_error("Windows Sockets init.",get_last_error());
        exit_app(EXIT_FAILURE);
    }
    if (local_server() || other_clients()) {
        myprintf(Red,"Another BotikKey client is running. Several clients should not run simultaneously.\n");
        do_sleep(2000);
        exit_app(EXIT_FAILURE);
    }
    if (strlen((char*)dpassword) != 0) {
        if (kchain) {
            if (!kchain_item_exists(dpassword)) {
                kchain_item_add(dpassword);
            }
        } else if (Kchain) {
            kchain_item_modify(dpassword);
        }
    }
    password = new_ustring(pwlen);
    if (!pwdfile && strlen((char*)dpassword)==0) {
        /* if no password file specified */
        if (getenv("BOTIKKEY")) {
            password = (ucharp)strncpy((char*)password,getenv("BOTIKKEY"),pwlen);
        } else {
            if (kchain) {
                if (!kchain_item_exists(dpassword)) {
                    query_password((char*)password,pwlen); /* read pswd from stdin */
                    make_digest(password,dpassword);
                    kchain_item_add(dpassword);
                }
            } else {
                query_password((char*)password,pwlen); /* read pswd from stdin */
            }
        }
    } else if (pwdfile && strlen((char*)dpassword)==0) {
        /* read the 1st line of the specified file */
        if (read_line(pwdfile,(char*)password,pwlen)) {
            free(password);
            return 1;
        }
    }
    if (strlen((char*)dpassword)==0) {
        make_digest(password,dpassword);
        if (kchain || Kchain) kchain_item_modify(dpassword);
    }
    free(password);
    return 0;
}
/***************** BotikKey protocol impl. **************************/
int main_loop(char *server, ucharp dpassword) {
    int n=0, ni=0, autos=1, sock, r, port=BKPORT, timeo=60;
    char *recv_buf, *challenge, *response, *tmptimeo, *chr,
        gateways[MAXGATES][IPLEN];
    time_t chtime;

    if (strlen(server) != 0) autos=0;
    tmptimeo = new_string(4);
    recv_buf = new_string(MSGSIZE);
    response = new_string(MSGSIZE);
    if (autos) {
        /* if no '-s' command-line option specified, then we have to probe
        several hosts for availability of network service on port 12040;
        these hosts include gateway(s) and some predefined Botik servers; */
        n+=find_gateways(gateways,3);
        add_gateway("192.168.73.31", gateways[n++]); /* grand-router.botik.ru */
        add_gateway("192.168.73.58", gateways[n++]); /* grand2-router.botik.ru */
        add_gateway("192.168.73.72", gateways[n++]); /* grand3-router.botik.ru */
    }
    reconnect:;
    if (autos) {
        if (ni < n) {
            server = gateways[ni++];
        } else {
            myprintf(Red,"No Key-server could be discovered. Please specify the correct address via the '-s' switch.\n");
            exit_app(EXIT_FAILURE);
        }
    }
    chtime = time(0);
    sock = connect_to(server,port);
    if (sock < 0) {
        if (verbose) myprintf(Red,"couldn't connect to server\n");
        close_socket(sock);
        do_sleep(3000);
        goto reconnect;
    }
    while (sock_readable(sock)) {
        if (time(0) - chtime > timeo+5) goto reconnect; /* should reconnect if no 'challenge' message received within 65 seconds */
        memset(recv_buf,0,MSGSIZE);
        r = recv(sock, recv_buf, MSGSIZE, 0);
        trim(recv_buf);
        if (r > 0) {
            if (strncmp(recv_buf,"BOTIKKEY",8) == 0) {
                if (verbose) myprintf(Yellow,"recv: %s\n", recv_buf);
                chtime = time(0);
                /* search for Timeout field */
                if ((chr = strchr(recv_buf,'=')) != NULL) {
                    strncpy(tmptimeo,++chr,3);
                    tmptimeo[3]='\0';
                    r=atoi(tmptimeo);
                    if (r > 0 && r <= 600) timeo=r;
                }
            } else if (strncmp(recv_buf,"CHALLENGE",9) == 0) {
                if (verbose) myprintf(Yellow,"recv: %s\n", recv_buf);
                chtime = time(0);
                challenge = strchr(recv_buf,' ')+1;
		        memset(response,0,MSGSIZE);
                get_response(challenge,(char*)dpassword,response);
                if (sock_writable(sock)) {
                    if (verbose) myprintf(Yellow,"send: %s",response);
                    r = send(sock,response,strlen(response),0);
                    if (r < 0) {
                        print_error("send",get_last_neterr());
                        close_socket(sock);
                        goto reconnect;
                    }
                } else {
                    close_socket(sock);
                    goto reconnect;
                }
                memset(recv_buf,0,MSGSIZE);
            } else if (strncmp(recv_buf,"GOOD",4) == 0) {
                autos=0;
                if (verbose) {
                    myprintf(Green,"recv: %s\n", recv_buf);
                    myprintf(Yellow,"waiting %d sec...\n",timeo);
                } else {
                    myprintf(Green,"%s\n", recv_buf);
                }
                goto wait;
            } else if (strncmp(recv_buf,"SORRY",5) == 0) {
                if (verbose) myprintf(Red,"recv: %s\n", recv_buf);
                if (!verbose) myprintf(Red,"%s\n", recv_buf);
                close_socket(sock);
                do_sleep(1000);
                goto reconnect;
            } else if (strncmp(recv_buf,"INFO",4) == 0 || strncmp(recv_buf,"MESSAGE",7) == 0) {
                if (verbose) myprintf(Green,"recv: %s\n", recv_buf);
                if (!verbose) myprintf(Green,"%s\n", recv_buf);
            } else {
                if (verbose) myprintf(Yellow,"recv: %s\n", recv_buf);
            }
        } else if (r < 0) {
            print_error("recv",get_last_neterr());
            close_socket(sock);
            do_sleep(1000);
            goto reconnect;
        }
        do_sleep(1000);
        continue;
        wait:;
        do_sleep(timeo*1000);
    }
    if (verbose) myprintf(Red,"lost connection\n");
    close_socket(sock);
    do_sleep(1000);
    goto reconnect;
}
/*******************************************/
void exit_app(int code) {
    unload_netlib();
    exit(code);
}
/*******************************************/
void signal_handler(int sig) {
    if (sig==SIGINT || sig==SIGTERM) {
        unload_netlib();
        exit(1);
    }
#ifndef _WIN32
    if (sig==SIGUSR1) strncpy(mode,"WORLD",6);
    if (sig==SIGUSR2) strncpy(mode,"LOCAL",6);
#endif
}
/*******************************************/
int main(int argc, char** argv) {
    int hashlen=2*MD5_HASHBYTES+1, srvlen=64, r;
    ucharp dpassword;
    char *server;

    srand((unsigned)get_tick_count());
    mode = new_string(6);
    strncpy(mode,"WORLD",6);
    signal(SIGINT,signal_handler);
    signal(SIGTERM,signal_handler);
#ifndef _WIN32
    signal(SIGUSR1,signal_handler);
    signal(SIGUSR2,signal_handler);
#endif
    set_title(APPNAME " " VER);
    server = new_string(srvlen);
    server[0]='\0';
    dpassword = new_ustring(hashlen);
    dpassword[0]='\0'; /* dpassword == MD5(bkey_password) */
    r=parse_args(argc,argv,server,srvlen,dpassword,hashlen);
    return r || main_loop(server,dpassword);
}
/*
  this code is best viewed fullscreen with FAR
*/
